/* 
 * Copyright (C) 1999-2000 Roberto Zunino
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <gtk/gtk.h>

#include <netdb.h>
#include <sys/poll.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>

#include <config.h>

#if defined(USE_ADNS)
#include <adns.h>
#elif defined (USE_ARES)
#include <sys/socket.h>
#include <ares.h>
#else
#error neither USE_ARES nor USE_ADNS defined!
#endif

#include "gnome-adns.h"

/* FIXME: portability
 *
 * poll() is not POSIX, but both this and (part of) adns require
 * struct pollfd to be defined
 *
 * I also convert from (struct GPollFD *) to (struct pollfd *) and vice versa
 *
 */

/* ChangeLog
 *
 * 17:16:00 CET 25 Nov 1999:   Roberto Zunino <zunrob@users.sourceforge.net>
 * First working version (alpha).
 * Using adns 0.6 (fixes memory leaks of 0.5).
 * User visible API could change in the future: you've been warned!
 * Simple test program (gnomeadnstest) written & working.
 * Used Memprof to fix memory leaks (seems ok now).
 *
 */

/* TODO
 *
 * Refine API
 * Documentation
 * Better adns/ares error handling
 * IPv6, but only when adns/ares IPv6 suppport has been added.
 * Should I use only one state for multiple queries to save memory?
 *
 * Note: this should work with both GNU adns and ares. This means that 
 * gnome-socket currently has _three_ different ways to do async dns
 * lookup. This also means that this file is full of #ifdef and friends.
 * I should drop adns support since it makes this file a bit hard to 
 * understand.
 */

/*
 * GnomeAdnsCallback:
 * @address: the result of the dns lookup. Its type is 
 * GNOME_NET_ADDRESS_INVALID iff the lookup failed. You haven't to destroy
 * (free) the address: the library will do this for you when you return from
 * the callback. If you need it after the callback returns you must
 * call #gnome_net_address_dup to obtain a safe copy.
 * @user_data: the pointer you passed to gnome_adns_lookup.
 */

#undef VERBOSE

/* --- source funcs for Glib --- */
static gboolean gnome_adns_prepare  (gpointer  source_data, 
				     GTimeVal *current_time,
				     gint     *timeout,
				     gpointer  user_data);
static gboolean gnome_adns_check    (gpointer  source_data,
				     GTimeVal *current_time,
				     gpointer  user_data);
static gboolean gnome_adns_dispatch (gpointer  source_data, 
				     GTimeVal *current_time,
				     gpointer  user_data);
static void gnome_adns_destroy      (gpointer data);

static GSourceFuncs gnome_adns_funcs = {
  gnome_adns_prepare,
  gnome_adns_check,
  gnome_adns_dispatch,
  gnome_adns_destroy
} ;

/* --- internal structures --- */
static GTree *tag_tree;
static gint next_tag;
static gboolean gnome_adns_initialized = FALSE;

typedef struct {
  /* user visible tag: 0 is _not_ a valid tag */
  guint tag;
  /* what kind of address the user wants? */
  GnomeNetAddressType addr_type;
  /* user callback and its user_data parameter */
  GnomeAdnsCallback cb;
  gpointer cb_data;

#if defined(USE_ADNS)
  /* adns stuff */
  adns_state adns__state;
  adns_query adns__query;
  adns_answer *adns__answer;
#else
  /* ares stuff */
  ares_channel ares__channel;
  /* ares__result is != NULL iff gnome_ares_callback has been called */
  GnomeNetAddress *ares__result; 
  /* ares__hostname is != NULL iff ares_gethostbyname() has _not_ been
   * called yet */
  gchar *ares__hostname; 
#endif

  /* g_main_add_poll stuff */
  GPollFD *fds;
  gint nfds;
  /* g_source_add tag : this is an *internal* (non user-visible) tag */
  guint source;
} GnomeAdnsStruct;

/* --- prototypes --- */
#if defined(USE_ADNS)
static void gnome_adns_beforepoll   (GnomeAdnsStruct *,
				     struct timeval *);
#else /* USE_ADNS -- now ares-related funcs */
static void gnome_ares_beforepoll   (GnomeAdnsStruct *);
static void gnome_ares_callback     (void *user_data, 
				     int status, 
				     struct hostent *host);
#endif

static void gnome_adns_init(void);
static gint gnome_adns_compare      (gconstpointer one, 
				     gconstpointer two);

/* --- user visible API ------------------- */

/*
 * gnome_adns_lookup:
 * @hostname: the name to be resolved.
 * @address_type: this must be GNOME_NET_ADDRESS_INET4 for now.
 * @callback: the function to call when dns lookup has been performed.
 * @callback_data: the data to pass to the @callback.
 *
 * Starts an asynchronous dns lookup. If you want, at any time you can call
 * #gnome_adns_abort to abort the lookup: in this case the  @callback will 
 * not be called.
 *
 * Returns: a valid tag or 0 in case of error.
 *
 */
gint gnome_adns_lookup(const gchar *hostname,
		       GnomeNetAddressType address_type,
		       GnomeAdnsCallback callback,
		       gpointer callback_data)
{
  GnomeAdnsStruct *s = NULL;

  gnome_adns_init();

  if (address_type != GNOME_NET_ADDRESS_INET4)
    {
      g_warning("gnome_adns_lookup: are you kidding? only IPv4 for now...");
      return 0;
    }

  /* initialization of the GnomeAdnsStruct */
  s = g_new(GnomeAdnsStruct, 1);
  s->tag = next_tag++;
  if (next_tag == 0) /* skip zero if roundover */
    next_tag = 1;
  s->addr_type = address_type;
  s->cb = callback;
  s->cb_data = callback_data;
  s->fds = NULL;
  s->nfds = 0;

#if defined(USE_ADNS)
  /* start the DNS lookup request */
  s->adns__state = NULL;
  s->adns__query = NULL;
  s->adns__answer = NULL;

  if (adns_init(&s->adns__state, adns_if_noautosys, 0) != 0)
    {
      g_warning("gnome_adns_lookup: adns_init != 0\n");
      return 0;
    }
  if (adns_submit(s->adns__state,
		  hostname,
		  adns_r_a,
		  adns_qf_owner | adns_qf_cname_loose,
		  NULL,
		  &s->adns__query) != 0)
    {
      g_warning("gnome_adns_lookup: adns_submit != 0\n");
      return 0;
    }

  gnome_adns_beforepoll(s, NULL);
#else /* USE_ADNS -- now ares */
  s->ares__channel = NULL;
  s->ares__result = NULL;
  s->ares__hostname = g_strdup(hostname);

  if (ares_init(&s->ares__channel) != ARES_SUCCESS)
    {
      g_warning("gnome_adns_lookup: ares_init != ARES_SUCCESS\n");
      return 0;
    }
#endif

  s->source = g_source_add(G_PRIORITY_DEFAULT,
			   FALSE, /* can_recurse */
			   &gnome_adns_funcs,
			   s,     /* source_data */
			   s,     /* user_data   */
			   NULL   /* gnome_adns_destroy */ ); 

  g_tree_insert(tag_tree, &s->tag, s);
  return s->tag;
}

/*
 * gnome_adns_abort:
 * @tag: the tag obtained from #gnome_adns_lookup.
 *
 * Aborts a dns lookup started by #gnome_adns_lookup.
 *
 * Returns: TRUE on success, FALSE otherwise.
 *
 */
gboolean gnome_adns_abort (gint tag)
{
  GnomeAdnsStruct *s = NULL;

#ifdef VERBOSE
  g_print("gnome_adns_abort(tag = %d)\n", tag);
#endif
  
  gnome_adns_init(); /* paranoia */

  s = (GnomeAdnsStruct *) g_tree_lookup(tag_tree, &tag);
  g_return_val_if_fail(s, FALSE);

#if defined(USE_ADNS)
  adns_cancel(s->adns__query);
  s->adns__query = NULL;
#endif
  
  /* g_source_remove will call gnome_adns_destroy which will do the real
     cleanup */
  if (! g_source_remove(s->source))
    g_warning("gnome_adns_abort(tag = %d): source %d not found!",
	      tag, s->source);

  return TRUE;
}

/* ----------------------------------------- */
static void gnome_adns_init(void)
{
  if (gnome_adns_initialized)
    return;
  gnome_adns_initialized = TRUE;
  next_tag = 1;
  tag_tree = g_tree_new(gnome_adns_compare);
}

static gboolean gnome_adns_prepare  (gpointer source_data, 
				     GTimeVal *current_time,
				     gint     *timeout,
				     gpointer user_data)
{
#if defined(USE_ARES) || defined(VERBOSE)
  GnomeAdnsStruct *s = (GnomeAdnsStruct *) source_data;
#endif

#ifdef VERBOSE
  g_print("gnome_adns_prepare tag=%d\n",
	  s->tag);
#endif

  /* This is here only because ares_gethostbyname could call the callback
   * immediately */
#if defined(USE_ARES)
  if (s->ares__hostname) /* this means the query has not been sent yet */
    {
      /* Note: this could call gnome_ares_callback immediately */
      ares_gethostbyname(s->ares__channel, s->ares__hostname,
			 AF_INET, gnome_ares_callback, s); /* FIXME: IPv4 */
      g_free(s->ares__hostname);
      s->ares__hostname = NULL;

      if (s->ares__result) /* gnome_ares_callback has been called */
	return TRUE;

      gnome_ares_beforepoll(s);
    }
#endif /* USE_ARES */

  *timeout = -1;
  return FALSE;
}

static gboolean gnome_adns_check    (gpointer source_data,
				     GTimeVal *current_time,
				     gpointer user_data)
{
  GnomeAdnsStruct *s = source_data;
  gint i;

#ifdef VERBOSE
  g_print("gnome_adns_check tag=%d\n", s->tag);
#endif

  for (i=0 ; i<s->nfds ; i++)
    if (s->fds[i].revents & s->fds[i].events)
      return TRUE;
  return FALSE;
}

#if defined(USE_ADNS)
static gboolean gnome_adns_dispatch (gpointer source_data, 
				     GTimeVal *current_time,
				     gpointer user_data)
{
  GnomeAdnsStruct *s = source_data;
  struct timeval now;

#ifdef VERBOSE
  g_print("gnome_adns_dispatch tag=%d\n", s->tag);
#endif /* VERBOSE */

  gettimeofday(&now, NULL);
  adns_afterpoll(s->adns__state,
		 (struct pollfd *) s->fds, 
		 s->nfds,
		 &now);
  
  if (adns_check(s->adns__state,
		 &s->adns__query,
		 &s->adns__answer,
		 NULL) == 0)
    {
      /* FIXME : find a better way .......... */
      GnomeNetAddress *addr = NULL;
      if (s->adns__answer->status == adns_s_ok && s->adns__answer->nrrs)
	{
	  guint32 a = ntohl(s->adns__answer->rrs.inaddr[0].s_addr); 
      
	  addr = gnome_net_address_new(s->addr_type);
	  g_memmove(addr->address, &a, addr->address_size);

#ifdef VERBOSE
	  g_print("nrrs = %d, answer == %d.%d.%d.%d\n",
		  s->adns__answer->nrrs,
		  (a & 0xff000000) >> 24,
		  (a & 0x00ff0000) >> 16,
		  (a & 0x0000ff00) >> 8,
		  a & 0x000000ff);
#endif /* VERBOSE */
	}
      else
	{
	  addr = gnome_net_address_new(GNOME_NET_ADDRESS_INVALID);

#ifdef VERBOSE
	  g_print("dns lookup failed....\n");
#endif /* VERBOSE */
	}

      if (s->cb)
	s->cb(addr, s->cb_data);

      gnome_net_address_destroy(addr);
      addr = NULL;
      return FALSE; /* remove source */
    }

  /* lookup is not complete, try again... */
  gnome_adns_beforepoll(s, &now);

  return TRUE; /* don't remove source */
}

#else /* USE_ADNS -- now the ares version: */
static gboolean gnome_adns_dispatch (gpointer source_data, 
				     GTimeVal *current_time,
				     gpointer user_data)
{
  GnomeAdnsStruct *s = source_data;
  fd_set read_fds, write_fds;
  gint i;

  /* Note: s->ares__result isn't NULL iff gnome_ares_callback has already been 
   * called, i.e. iff ares_gethostbyname called it immediately */
  if (! s->ares__result)
    { /* we must process the data from the net */
      /* poll2select */
      FD_ZERO(&read_fds);
      FD_ZERO(&write_fds);
      for (i = 0; i < s->nfds ; i++)
	{
	  if (s->fds[i].revents | G_IO_IN)
	    FD_SET(s->fds[i].fd, &read_fds);
	  if (s->fds[i].revents | G_IO_OUT)
	    FD_SET(s->fds[i].fd, &write_fds);
	}
      /* this could set s->ares__result */
      ares_process(s->ares__channel, &read_fds, &write_fds);
    }

  /* test if dns lookup is finished */
  if (s->ares__result)
    {
      if (s->cb)
	s->cb(s->ares__result, s->cb_data);
      
      gnome_net_address_destroy(s->ares__result);
      s->ares__result = NULL;
      return FALSE; /* remove source */
    }

  /* lookup is not complete */
  gnome_ares_beforepoll(s);
  return TRUE; /* don't remove source */
}

#endif /* adns / ares for gnome_adns_dispatch */

static void gnome_adns_destroy (gpointer source_data)
{
  GnomeAdnsStruct *s = source_data;
  gint i;

#ifdef VERBOSE
  g_print("gnome_adns_destroy() tag=%d\n",s->tag);
#endif

  g_return_if_fail(s);

  /* DO NOT remove the source here! It's already handled by glib */
  for (i=0 ; i<s->nfds ; i++)
    g_main_remove_poll(s->fds + i);

  g_tree_remove(tag_tree, &s->tag);

#if defined (USE_ADNS)
  adns_finish(s->adns__state);
  s->adns__state = NULL;

  /* from adns 0.5 docs : Answers are malloc'd as a single piece of memory; 
     pointers in the answer struct point into further memory in the answer.*/
  if (s->adns__answer)
    free(s->adns__answer);
  s->adns__answer = NULL;

#else /* USE_ADNS -- now ares */
  ares_destroy(s->ares__channel);
  s->ares__channel = NULL;
#endif

  /* FIXME : other things to free() ? */

  /* don't free the query! -- adns doesn't like it */
  /* adns_cancel(s->adns__query);
     s->adns__query = NULL; */

  if (s->fds)
    g_free(s->fds);
  s->fds = NULL;

  g_free(s);
}

/* --- utility functions ----------------------- */

static gint gnome_adns_compare (gconstpointer one, 
				gconstpointer two)
{
  return (*(gint *) one) - (*(gint *) two);
}

#if defined(USE_ADNS)
/* if the caller has already done gettimeofday() can pass the result in 
 * "now", if not it should pass NULL as "now" */
static void gnome_adns_beforepoll(GnomeAdnsStruct *s,
				  struct timeval *now)
{
  struct timeval time;
  gint i;

  g_return_if_fail(s);

#ifdef VERBOSE
  g_print("gnome_adns_beforepoll: tag = %d\n", s->tag);
#endif

  if (! now)
    {
      gettimeofday(&time, NULL);
      now = &time;
    }

  /* s->fds is NULL iff gnome_adns_beforepoll has been called from 
   * gnome_adns_lookup */
  if (s->fds)
    { /* remove old fds */
      for (i=0 ; i<s->nfds ; i++)
	g_main_remove_poll(s->fds + i);
      g_free(s->fds);
      s->fds = NULL;
      s->nfds = 0;
    }

  adns_beforepoll(s->adns__state,
		  NULL,
		  &s->nfds,
		  NULL,
		  now); /* get nfds */
  s->fds = g_new(GPollFD, s->nfds);

  if (adns_beforepoll(s->adns__state,
		      (struct pollfd *) s->fds, 
		      &s->nfds, 
		      NULL,
		      now) != 0)
    g_warning("gnome_adns_beforepoll: adns_beforepoll failed!!!!!!!");

  for (i=0 ; i<s->nfds ; i++)
    g_main_add_poll(s->fds + i, G_PRIORITY_DEFAULT);
}

#else /* USE_ADNS -- now ares */

static void gnome_ares_beforepoll   (GnomeAdnsStruct *s)
{
  fd_set read_fds, write_fds;
  gint nfds, real_nfds, i, j;

#ifdef VERBOSE
  g_print("gnome_ares_beforepoll: tag = %d\n", s->tag);
#endif

  FD_ZERO(&read_fds);
  FD_ZERO(&write_fds);
  nfds = ares_fds(s->ares__channel, &read_fds, &write_fds);

  if (nfds == 0)
    g_warning("gnome_ares_beforepoll: nfds == 0!\n");

  /* free the old structures, if any */
  if (s->fds)
    {
      for (i=0 ; i<s->nfds ; i++)
	g_main_remove_poll(s->fds + i);
      g_free(s->fds);
      s->fds = NULL;
      s->nfds = 0;
    }

  /* count the present fds */
  real_nfds = 0;
  for (i = 0; i < nfds; i++)
    if (FD_ISSET(i, &read_fds) || FD_ISSET(i, &write_fds))
      real_nfds++;

  s->nfds = real_nfds;
  s->fds = g_new(GPollFD, s->nfds);
  for (i = 0, j = 0; i<nfds ; i++)
    if (FD_ISSET(i, &read_fds) || FD_ISSET(i, &write_fds))
      {
	s->fds[j].fd = i;
	s->fds[j].events = s->fds[j].revents = 0;
	if (FD_ISSET(i, &read_fds)) 
	  s->fds[j].events |= G_IO_IN;
	if (FD_ISSET(i, &write_fds)) 
	  s->fds[j].events |= G_IO_OUT;
	g_main_add_poll(s->fds + j, G_PRIORITY_DEFAULT);
	j++;
      }
}

static void gnome_ares_callback     (void *user_data, 
				     int status, 
				     struct hostent *host)
{
  GnomeAdnsStruct *s = (GnomeAdnsStruct *) user_data;

#ifdef VERBOSE
  g_print("gnome_ares_callback: tag = %d\n", s->tag);
#endif

  switch (status)
    {
    case ARES_EDESTRUCTION:
      break; /* ignore */
    case ARES_SUCCESS:
      if (! host)
	/* dns lookup failed */
	s->ares__result = gnome_net_address_new(GNOME_NET_ADDRESS_INVALID);
      else
	{
	  guint32 a = ntohl(*(guint32 *) host->h_addr_list[0]); 

	  s->ares__result = gnome_net_address_new(s->addr_type);
	  g_memmove(s->ares__result->address, &a, 
		    s->ares__result->address_size);

#ifdef VERBOSE
	  g_print("answer == %d.%d.%d.%d\n",
		  (a & 0xff000000) >> 24,
		  (a & 0x00ff0000) >> 16,
		  (a & 0x0000ff00) >> 8,
		  a & 0x000000ff);
#endif /* VERBOSE */
	}
      break;
    case ARES_ENOTFOUND:
    case ARES_EBADNAME:
      s->ares__result = gnome_net_address_new(GNOME_NET_ADDRESS_INVALID);
      break;
    default:
      {
	char *errmem;
	g_warning("gnome_ares_callback: ares error %s\n",
		  ares_strerror(status, &errmem));
	ares_free_errmem(errmem);
	break;
      }
    }
  return;
}

#endif
