/* 
 * Copyright (C) 1999-2000 Roberto Zunino
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <gtk/gtk.h>

#include <sys/types.h>
#include <sys/socket.h> 
#include <netinet/in.h> 
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>

#include <config.h>

#if defined(USE_ADNS) || defined(USE_ARES)
#include "gnome-adns.h"
#else /* USE_GNOME instead */
#include "libgnome/gnome-defs.h"
#include "libgnomeui/gnome-dns.h"
#endif

#include "gnome-socket.h"

#undef VERBOSE

/* misc. private utilities & callbacks */
static void gnome_socket_initialize_first_time(void);
#if defined(USE_GNOME)
static void gnome_socket_dns_callback (guint32 ip_addr, void *data);
#endif
static void gnome_socket_adns_callback (GnomeNetAddress *addr,
					gpointer user_data);
static gboolean gnome_socket_callback(GIOChannel *channel,
				      GIOCondition condition,
				      gpointer data);
static gint gnome_socket_compare (gconstpointer one, 
				  gconstpointer two);
static gboolean gnome_socket_proto_translate(GnomeNetProtocol proto,
					     gint *domain, gint *type, 
					     gint *protocol);

typedef enum {
  STATUS_RESOLVING    = 0,  /* waiting for dns lookup */
  STATUS_RESOLVED     = 1,  /* dns lookup done */
  STATUS_CONNECTING   = 2,  /* waiting for connect() to terminate */
  STATUS_CONNECTED    = 3,  /* channel ready for I/O */
  STATUS_LISTENING    = 4,  /* waiting for an incoming call */
  STATUS_REFUSED      = 5,  /* callback closed connection in a FYI call:
			     * i.e. : GNOME_SOCKET_JUST_RESOLVED */
  STATUS_ERROR        = 6   /* generic error (e.g. bind()==-1) */
} CommStatus;


typedef struct 
{
  /* user-visible tag (the "invalid" tag 0 should not appear here) */
  guint tag;
  /* socket status: see macros above */
  CommStatus status;
  /* addr of the _other_ end, or NULL if STATUS_LISTENING */
  GnomeNetAddress *addr;
  GnomeNetProtocol proto;
  /* port of the _other_ end, or local port if STATUS_LISTENING */
  guint port;
  /* callback for general communication: see docs for semantics */
  GnomeSocketCallback cb;
  gpointer cb_data;
  /* the comm channel */
  GIOChannel *io_channel;
  /* tag from g_io_add_watch(io_channel,...), or tag from 
   * gnome_(a)dns_lookup(...) if STATUS_RESOLVING */
  guint source;
} GnomeSocket;

static gboolean gnome_socket_initialized = FALSE;
static GTree *gnome_socket_tree = NULL; /* AVL tree for GnomeSocket */
static guint next_tag = 1;

/* FIXME : general portability?? */
/* FIXME : check for interrupted system calls (EINTR) */
/* FIXME : really works only for IPV4 and TCP ... */

/*
 *  gnome_socket_connect:
 *  @hostname: the name of the host to contact.
 *  @addr_type: this must be GNOME_NET_ADDRESS_INET4 for now.
 *  @proto: the network protocol (like GNOME_NET_PROTOCOL_TCP or
 *  GNOME_NET_PROTOCOL_UDP).
 *  @port: the destination port (if applicable to the protocol).
 *  @callback: the function to call when dns lookup has been performed, when
 *  connection has been established (or a network error occoured), when 
 *  data has been received from the net and when the connection has been
 *  closed. See #GnomeSocketCallback for more info.
 *  @callback_data: the user data that will be passed whenever the @callback
 *  is called.
 *
 *  Contacts a remote host at a specified port, calling the @callback to
 *  notify anything interesting.
 *
 *  Please note: if gnome-socket was compiled with --with-resolver-lib=gnome
 *  this function could call the @callback immediately in certain cases.
 *
 *  Returns: a gnome_socket tag (0 in case of error).
 */
guint
gnome_socket_connect (const char *hostname, 
		      GnomeNetAddressType addr_type,
		      GnomeNetProtocol proto, guint port,
		      GnomeSocketCallback callback,
		      gpointer callback_data)
/* Note on semantics:
 * it could call __immediately__ the callback if (dns cache hit &&
 * connection refused)    [sigh!] 
 * If USE_ADNS or USE_ARES this can't happen.
 */
{
  GnomeSocket *sock;
  guint temp_source, assigned_tag;

  g_return_val_if_fail(callback != NULL, 0);
  g_return_val_if_fail(hostname != NULL, 0);
  g_return_val_if_fail(gnome_socket_proto_translate(proto,NULL,NULL,NULL), 0);

  gnome_socket_initialize_first_time();
  sock = g_new(GnomeSocket, 1);
  sock->tag = assigned_tag = next_tag++;
  if (next_tag == 0)
    next_tag = 1;
  sock->status = STATUS_RESOLVING;
  sock->addr = NULL;
  sock->proto = proto;
  sock->port = port;
  sock->cb = callback;
  sock->cb_data = callback_data;
  sock->io_channel = NULL;
  sock->source = 0;

  g_tree_insert(gnome_socket_tree,
		&(sock->tag),
		sock);

#if defined(USE_ADNS) || defined(USE_ARES)
  temp_source = gnome_adns_lookup(hostname,
				  addr_type,
				  gnome_socket_adns_callback,
				  sock);
#else /* USE_GNOME instead */
  temp_source = gnome_dns_lookup(hostname,
				 gnome_socket_dns_callback,
				 sock);
#endif

  /* Only ifndef USE_ADNS or USE_ARES:
   *
   * gnome_dns_lookup could immediately call the callback that set sock->source
   * to the socket: if this happens we should not override it [sigh]
   * In this case, temp_source is 0.
   *
   * Beware! Don't rely that sock points to a valid area here unless 
   * temp_source!=0 User might have called gnome_socket_close! 
   */

  if (temp_source != 0) 
    sock->source = temp_source;

  return assigned_tag; /* == sock->tag */
}

/*
 * gnome_socket_listen:
 * @proto: the network protocol (like GNOME_NET_PROTOCOL_TCP or
 *  GNOME_NET_PROTOCOL_UDP).
 * @port: the local port (if applicable to the protocol).
 * @callback: the function to call when there's an incoming connection (or
 * a network error occoured), when data has been received from the net and
 * when the connection has been closed. See #GnomeSocketCallback for more info.
 * @callback_data: the user data that will be passed whenever the @callback
 *  is called.
 * @info: this must point to a valid allocated zone, and it's used to report
 * errors to the caller. If @info is not GNOME_SOCKET_OK it identifies what
 * system call failed. 
 * @error_no: this must point to a valid allocated zone, and it's equal to libc
 * errno. If @info is not GNOME_SOCKET_OK this identifies why the system call
 * failed.
 *
 * Opens a socket awaiting for incoming connections on a port.
 *
 * Returns: a gnome_socket tag (0 in case of error).
 */
guint
gnome_socket_listen(GnomeNetProtocol proto, guint port,
		    GnomeSocketCallback callback,
		    gpointer callback_data,
		    GnomeSocketInfo *info,
		    gint *error)
{
  GnomeSocket *sock;
  struct sockaddr_in s;
  gint sock_fd, flags;
  gint domain, type, protocol; /* for socket() */
  gint value; /* for setsockopt */

  g_return_val_if_fail(callback != NULL, 0);
  g_return_val_if_fail(info != NULL, 0);
  g_return_val_if_fail(error != NULL, 0);

  gnome_socket_initialize_first_time();
  *info = GNOME_SOCKET_OK;
  *error = 0;

  if (!gnome_socket_proto_translate(proto, &domain, &type, &protocol))
    { /* invalid proto */
      g_warning("gnome_socket_listen: invalid protocol");
      return 0;
    }

  if ((sock_fd = socket(domain, type, protocol)) == -1)
    {
      *info = GNOME_SOCKET_SOCKET_ERROR;
      *error = errno;
      return 0;
    }
  /* without REUSEADDR, closing a listening socket makes bind() to the same
   * port fail for the following two miutes */
  value = 1;
  if (setsockopt(sock_fd, SOL_SOCKET, SO_REUSEADDR,
		 &value, sizeof(value)) == -1)
    {
      g_warning("gnome_socket_listen: cannot set the REUSEADDR option");
      /* this isn't a fatal error, therefore don't return 0 */
    }
  /* set non-blocking fd */
  if ( ( (flags = fcntl(sock_fd, F_GETFL)) == -1) ||
       ( fcntl(sock_fd, F_SETFL, flags | O_NONBLOCK) == -1) ) 
    {
      close(sock_fd);
      *info = GNOME_SOCKET_FCNTL_ERROR;
      *error = errno;
      return 0;
    }
  
  s.sin_family = AF_INET; /* IPv4 */
  s.sin_port   = htons(port);
  s.sin_addr.s_addr = htonl(INADDR_ANY);   /* allow anyone to connect  */

  if (bind(sock_fd,
	   (struct sockaddr *) &s,
	   sizeof(struct sockaddr)) ==-1)
    {
      close(sock_fd);
      *info = GNOME_SOCKET_BIND_ERROR;
      *error = errno;
      return 0;
    }
  if (listen(sock_fd,10)==-1) /* FIXME: backlog == 10 may be insuff. */
    {
      close(sock_fd);
      *info = GNOME_SOCKET_LISTEN_ERROR;
      *error = errno;
      return 0;
    }

  sock = g_new(GnomeSocket, 1);
  sock->tag = next_tag++;
  if (next_tag == 0)
    next_tag = 1;
  sock->status = STATUS_LISTENING;
  sock->addr = NULL;
  sock->proto = proto;
  sock->port = port;
  sock->cb = callback;
  sock->cb_data = callback_data;
  sock->io_channel = NULL;
  sock->source = 0;

  g_tree_insert(gnome_socket_tree,
		&(sock->tag),
		sock);

  /* waiting for incoming connections */
  sock->io_channel = g_io_channel_unix_new(sock_fd);
  sock->source = g_io_add_watch(sock->io_channel,
				G_IO_IN | G_IO_ERR,
				(GIOFunc) gnome_socket_callback,
				sock);

  return sock->tag;
}

/*
 * gnome_socket_get_addr:
 * @tag: a gnome_socket tag created by #gnome_socket_connect or 
 * #gnome_socket_listen.
 *
 * Returns: the address of the other end or NULL if not applicable (e.g.
 * the socket is listening). The caller must call gnome_net_address_destroy
 * when it isn't needed anymore.
 */
GnomeNetAddress *
gnome_socket_get_addr(guint tag)
{
  GnomeSocket *sock;

  g_return_val_if_fail(tag > 0, NULL);
  gnome_socket_initialize_first_time(); /* paranoia */

  sock = g_tree_lookup(gnome_socket_tree, &tag);
  if (sock == NULL)
    {
      g_warning("gnome_socket_get_addr: tag %d not found!\n",tag);
      return NULL;
    }

  switch (sock->status)
    {
    case STATUS_ERROR:
      g_warning("gnome_socket_get_addr: you should already have "
		"received an error notification for tag %d!\n",tag);
      return NULL;
    case STATUS_LISTENING:
      g_warning("gnome_socket_get_addr: the socket for tag %d is "
		"a listening socket!\n",tag);
      return NULL;
    case STATUS_RESOLVING:
      g_warning("gnome_socket_get_addr: the socket for tag %d is "
		"still resolving!\n",tag);
      return NULL;
    case STATUS_REFUSED:
      g_warning("gnome_socket_get_addr: the socket for tag %d has "
		"already been closed!\n",tag);
      return NULL;
    case STATUS_RESOLVED:
    case STATUS_CONNECTING:
    case STATUS_CONNECTED:
      return gnome_net_address_dup(sock->addr);
    default:
      g_warning("gnome_socket_get_addr: sock->status == %d\n",sock->status);
      return NULL;
    }
}

/*
 * gnome_socket_get_port:
 * @tag: a gnome_socket tag created by #gnome_socket_connect or 
 * #gnome_socket_listen.
 *
 * Returns: if the socket is listening it returns the local port, if the
 * socket is connecting/ed it returns the remote port. Currently this 
 * works only for outcoming connections. It returns 0 if not applicable.
 */
guint
gnome_socket_get_port(guint tag)
{
  GnomeSocket *sock;

  g_return_val_if_fail(tag > 0, 0);
  gnome_socket_initialize_first_time(); /* paranoia */

  sock = g_tree_lookup(gnome_socket_tree, &tag);
  if (sock == NULL)
    {
      g_warning("gnome_socket_get_port: tag %d not found!\n",tag);
      return 0;
    }

  switch (sock->status)
    {
    case STATUS_ERROR:
      g_warning("gnome_socket_get_port: you should already have "
		"received an error notification for tag %d!\n",tag);
      return 0;
    case STATUS_REFUSED:
      g_warning("gnome_socket_get_port: tag %d already closed!\n",tag);
      return 0;
    case STATUS_RESOLVING:
    case STATUS_RESOLVED:
    case STATUS_LISTENING:
    case STATUS_CONNECTING:
    case STATUS_CONNECTED:
      return sock->port;
    default:
      g_warning("gnome_socket_get_port: sock->status == %d\n",sock->status);
      return 0;
    }
}

GnomeSocketCallback
gnome_socket_get_callback(guint tag)
{
  GnomeSocket *sock;

  g_return_val_if_fail(tag > 0, NULL);
  gnome_socket_initialize_first_time(); /* paranoia */

  sock = g_tree_lookup(gnome_socket_tree, &tag);
  if (sock == NULL)
    {
      g_warning("gnome_socket_get_callback: tag %d not found!\n",tag);
      return NULL;
    }

  switch (sock->status)
    {
    case STATUS_RESOLVING:
    case STATUS_RESOLVED:
    case STATUS_CONNECTING:
    case STATUS_CONNECTED:
    case STATUS_LISTENING:
    case STATUS_ERROR:
      return sock->cb;
    case STATUS_REFUSED:
      g_warning("gnome_socket_get_callback: tag %d already closed!\n",tag);
      return NULL; /* the socket has been closed */
    default:
      g_warning("gnome_socket_get_callback: sock->status == %d\n",
		sock->status);
      return NULL;
    }
}

gboolean
gnome_socket_set_callback(guint tag, GnomeSocketCallback callback)
{
  GnomeSocket *sock;

  g_return_val_if_fail(tag > 0, FALSE);
  gnome_socket_initialize_first_time(); /* paranoia */

  sock = g_tree_lookup(gnome_socket_tree, &tag);
  if (sock == NULL)
    {
      g_warning("gnome_socket_set_callback: tag %d not found!\n",tag);
      return FALSE;
    }

  switch (sock->status)
    {
    case STATUS_RESOLVING:
    case STATUS_RESOLVED:
    case STATUS_CONNECTING:
    case STATUS_CONNECTED:
    case STATUS_LISTENING:
    case STATUS_ERROR:
       sock->cb = callback;
       return TRUE;
    case STATUS_REFUSED:
      g_warning("gnome_socket_set_callback: tag %d already closed!\n",tag);
      return FALSE; /* the socket has been closed */
    default:
      g_warning("gnome_socket_set_callback: sock->status == %d\n",
		sock->status);
      return FALSE;
    }
}

gpointer
gnome_socket_get_callback_data(guint tag)
{
  GnomeSocket *sock;

  g_return_val_if_fail(tag > 0, NULL);
  gnome_socket_initialize_first_time(); /* paranoia */

  sock = g_tree_lookup(gnome_socket_tree, &tag);
  if (sock == NULL)
    {
      g_warning("gnome_socket_get_callback_data: tag %d not found!\n",tag);
      return NULL;
    }

  switch (sock->status)
    {
    case STATUS_RESOLVING:
    case STATUS_RESOLVED:
    case STATUS_CONNECTING:
    case STATUS_CONNECTED:
    case STATUS_LISTENING:
    case STATUS_ERROR:
      return sock->cb_data;
    case STATUS_REFUSED:
      g_warning("gnome_socket_get_callback_data: tag %d already closed!\n",
		tag);
      return NULL; /* the socket has been closed */
    default:
      g_warning("gnome_socket_get_callback_data: sock->status == %d\n",
		sock->status);
      return NULL;
    }
}

gboolean
gnome_socket_set_callback_data(guint tag, gpointer callback_data)
{
  GnomeSocket *sock;

  g_return_val_if_fail(tag > 0, FALSE);
  gnome_socket_initialize_first_time(); /* paranoia */

  sock = g_tree_lookup(gnome_socket_tree, &tag);
  if (sock == NULL)
    {
      g_warning("gnome_socket_set_callback_data: tag %d not found!\n",tag);
      return FALSE;
    }

  switch (sock->status)
    {
    case STATUS_RESOLVING:
    case STATUS_RESOLVED:
    case STATUS_CONNECTING:
    case STATUS_CONNECTED:
    case STATUS_LISTENING:
    case STATUS_ERROR:
       sock->cb_data = callback_data;
       return TRUE;
    case STATUS_REFUSED:
      g_warning("gnome_socket_set_callback_data: tag %d already closed!\n",
		tag);
      return FALSE; /* the socket has been closed */
    default:
      g_warning("gnome_socket_set_callback_data: sock->status == %d\n",
		sock->status);
      return FALSE;
    }
}

/*
 * gnome_socket_close:
 * @tag: a gnome_socket tag created by #gnome_socket_connect or 
 * #gnome_socket_listen.
 *
 * Closes an active connection or stops listening for incoming connections.
 *
 * Returns: TRUE on success, FALSE otherwise.
 */
gboolean
gnome_socket_close(guint tag)
{
  GnomeSocket *sock;

  g_return_val_if_fail(tag > 0, FALSE);
  gnome_socket_initialize_first_time(); /* paranoia */

  sock = g_tree_lookup(gnome_socket_tree, &tag);
  if (sock == NULL)
    {
      g_warning("gnome_socket_close: tag %d not found!\n",tag);
      return FALSE;
    }

#ifdef VERBOSE
  g_message("gnome_socket_close: sock={"
	    "tag=%d,source=%d,status=%d,...}",
	    sock->tag,sock->source,sock->status);
#endif  

  switch (sock->status)
    {
    case STATUS_RESOLVING:
#if defined(USE_ADNS) || defined(USE_ARES)
      gnome_adns_abort(sock->source);
#else
      gnome_dns_abort(sock->source);
#endif
      break;
    case STATUS_RESOLVED:
      sock->status = STATUS_REFUSED;
      return TRUE; /* we cannot free stuff now since close has been called 
		    * by a FYI callback */
    case STATUS_REFUSED:
      g_warning("gnome_socket_close: tag %d already closed!\n",tag);
      return FALSE; /* don't allow the FYI callback to insist */
    case STATUS_LISTENING:
    case STATUS_CONNECTING:
    case STATUS_CONNECTED:
      g_source_remove(sock->source);
      /* falltrough */
    case STATUS_ERROR:
      /* sock->io_channel could be NULL if DNS lookup failed */
      if (sock->io_channel != NULL) 
	{
	  g_io_channel_close(sock->io_channel);
	  g_io_channel_unref(sock->io_channel);
	}
      break;
    default:
      g_warning("gnome_socket_close: sock->status == %d\n",sock->status);
      break;
    }
  
  g_tree_remove(gnome_socket_tree, &tag);

  if (sock->addr)
    {
      gnome_net_address_destroy(sock->addr);
      sock->addr = NULL;
    }
  g_free(sock);
  sock = NULL;

  return TRUE;
}

/* -------------------------------------------------------- */

#if defined(USE_GNOME)
/*
 * Don't reinvent the weel here: all the work is done by 
 * gnome_socket_adns_callback. Reuse it: it's safer for internal coherence and
 * FYI calls handling.
 */
static void
gnome_socket_dns_callback (guint32 ip_addr, void *data)
{ 
  GnomeNetAddress *addr = NULL;

  if (ip_addr == 0) /* dns lookup failed */
      addr = gnome_net_address_new(GNOME_NET_ADDRESS_INVALID);
  else
    {
      addr = gnome_net_address_new(GNOME_NET_ADDRESS_INET4);
      *(guint32 *) addr->address = ip_addr; 
    }

  gnome_socket_adns_callback(addr, data);
  gnome_net_address_destroy(addr);
}
#endif

static void
gnome_socket_adns_callback (GnomeNetAddress *addr, gpointer data)
{
  GnomeSocket *sock = (GnomeSocket *) data;
  gint sock_fd, flags;
  gint domain, type, protocol; /* for socket() */
  struct sockaddr_in s;
  
  g_return_if_fail(data != NULL);
  g_return_if_fail(addr != NULL);

#ifdef VERBOSE
  g_message("gnome_socket_adns_callback (...,sock={"
	    "tag=%d,source=%d,status=%d,...})",
	    sock->tag,sock->source,sock->status);
#endif
  g_print("&addr == %p\n", &addr);

  if (addr->type == GNOME_NET_ADDRESS_INVALID)
    { /* DNS failed */
      sock->status = STATUS_ERROR;
      sock->cb(sock->tag,
	       GNOME_SOCKET_DNS_ERROR,
	       0,       /* extra_info */
	       NULL,    /* IOChannel */
	       sock->cb_data);
      return;
    }  

  sock->addr = gnome_net_address_dup(addr);

  /* FIXME : cb -may-> socket_close -> crash 
   * should be fixed now ... watch out, anyway */
  sock->status = STATUS_RESOLVED;
  sock->cb(sock->tag,
	   GNOME_SOCKET_JUST_RESOLVED,
	   0,       /* extra_info */
	   NULL,    /* IOChannel */
	   sock->cb_data);

  if (sock->status == STATUS_REFUSED)
    { /* cb has called close: we do it now*/
      sock->status = STATUS_ERROR;
      gnome_socket_close(sock->tag);
      return;
    }

  if (!gnome_socket_proto_translate(sock->proto, &domain, &type, &protocol))
    {
      g_warning("gnome_socket_adns_callback: invalid protocol?\n"
		"please report to the maintainers of gnome-socket "
		"this error.\n");
      sock->status = STATUS_ERROR;
      sock->cb(sock->tag,
	       GNOME_SOCKET_INVALID_PROTO,
	       0,
	       NULL,
	       sock->cb_data);
      return;
    }

  if ((sock_fd = socket(domain, type, protocol)) == -1)
     {
      sock->status = STATUS_ERROR;
      sock->cb(sock->tag,
	       GNOME_SOCKET_SOCKET_ERROR,
	       errno,
	       NULL,
	       sock->cb_data);
      return;
    }
  /* set non-blocking fd */
  if ( ( (flags = fcntl(sock_fd, F_GETFL)) == -1) ||
       ( fcntl(sock_fd, F_SETFL, flags | O_NONBLOCK) == -1) ) 
    {
      close(sock_fd);
      sock->status = STATUS_ERROR;
      sock->cb(sock->tag,
	       GNOME_SOCKET_FCNTL_ERROR,
	       errno,
	       NULL,
	       sock->cb_data);
      return;
    }

  s.sin_family      = AF_INET;
  s.sin_port        = htons(sock->port);
  s.sin_addr.s_addr = htonl(* (guint32 *) sock->addr->address);

  sock->status = STATUS_CONNECTING;
  if (connect(sock_fd,
	      (struct sockaddr *) &s,
	      sizeof(struct sockaddr)) == -1 &&
      errno != EINPROGRESS)
    {
      /* comm error */
      close(sock_fd);
      sock->status = STATUS_ERROR;
      sock->cb(sock->tag,
	       GNOME_SOCKET_CONNECT_ERROR,
	       errno,   
	       NULL,
	       sock->cb_data);
      return;
    }

  /* wait for connect() to finish, even if connect(...)==0 
   * i.e. connect() has already terminated: this is slightly slower
   * but avoids a special case. 
   * Moreover gnome_dns_lookup() can call
   * this function immediately [sigh!] and we can have something 
   * like gnome_socket_connect -> gnome_socket_dns_callback -> (user callback)
   * -> gnome_socket_close and SIGSEGV in gnome_socket_connect. 
   * Of course, this isn't possible ifdef USE_ADNS or USE_ARES
   */
  sock->io_channel = g_io_channel_unix_new(sock_fd);
  sock->source = g_io_add_watch(sock->io_channel,
				G_IO_OUT | G_IO_ERR,
				(GIOFunc) gnome_socket_callback,
				sock);
}

static gboolean
gnome_socket_callback(GIOChannel *channel,
		      GIOCondition condition,
		      gpointer data)
{
  GnomeSocket *sock = (GnomeSocket *) data;
  int result,sock_fd;
  size_t len;

  g_return_val_if_fail(sock != NULL, FALSE);
  g_return_val_if_fail(sock->cb != NULL, FALSE);
  g_return_val_if_fail(channel != NULL, FALSE);
  
#ifdef VERBOSE
  g_message("gnome_socket_callback: sock={"
	    "tag=%d,source=%d,status=%d,...}",
	    sock->tag,sock->source,sock->status);
#endif

  sock_fd = g_io_channel_unix_get_fd(channel);

  switch (sock->status)
    {
    case STATUS_CONNECTING:
      len = sizeof(result);
      if ( getsockopt(sock_fd, SOL_SOCKET, SO_ERROR,
		      &result, &len) == -1 ||
	   len != sizeof(result) )
	{
	  sock->status = STATUS_ERROR;
	  /* should we call sock->cb ?? */
	  g_warning("gnome_socket_callback: getsockopt error=%d",
		    errno);
	  return FALSE; /* remove from event loop */
	}
      
      if (result != 0)
	{
	  if (result == EINPROGRESS)
	    {  /* paranoia : is this really necessary?? */
	      g_warning("getsockopt can be EINPROGRESS !!!!");
	      return TRUE;
	    }
	  sock->status = STATUS_ERROR;
	  sock->cb(sock->tag,
		   GNOME_SOCKET_CONNECT_ERROR,
		   result,
		   NULL,
		   sock->cb_data);
	  return FALSE; /* remove from event loop */
	}

      sock->status = STATUS_CONNECTED;
      /* don't wait for connect() anymore */
      g_source_remove(sock->source); 
      /* wait for incoming data instead */
      sock->source = g_io_add_watch(channel,
				    G_IO_ERR | G_IO_IN,
				    (GIOFunc) gnome_socket_callback,
				    sock);
      /* call sock->cb to give it a chance to send something when connection is
       * ready */
      sock->cb(sock->tag,
	       GNOME_SOCKET_JUST_CONNECTED,
	       sock->tag, /* for coherence with gnome_socket_listen */
	       channel,
	       sock->cb_data);
      break;
      
    case STATUS_CONNECTED:
      /* simple proxy */
      sock->cb(sock->tag,
	       GNOME_SOCKET_OK,
	       0,
	       channel,
	       sock->cb_data);
      break;

    case STATUS_LISTENING:
      {
	GnomeSocket *remote_sock = NULL;
	gint remote_sock_fd;
	struct sockaddr_in remote_s;

	len = sizeof(struct sockaddr_in);
	remote_sock_fd = accept(sock_fd,
				(struct sockaddr *) &remote_s,
				&len);
	if (remote_sock_fd == -1)
	  {
	    /* paranoia check */
	    if (errno == EWOULDBLOCK)
	      return TRUE;
	    
	    sock->status = STATUS_ERROR;
	    sock->cb(sock->tag,
		     GNOME_SOCKET_ACCEPT_ERROR,
		     errno,
		     NULL,
		     sock->cb_data);
	    return FALSE; /* FIXME : remove from event loop? */	  
	  }
	
	/* DON'T remove the channel used to wait incoming connections 
	   according to the new semantics
	   g_source_remove(sock->source);
	   g_io_channel_close(channel);
	   g_io_channel_unref(channel);
	*/

	remote_sock = g_new(GnomeSocket, 1);
	remote_sock->tag = next_tag++;
	if (next_tag == 0)
	  next_tag = 1;
	remote_sock->port = ntohs(remote_s.sin_port);
	remote_sock->addr = gnome_net_address_new(GNOME_NET_ADDRESS_INET4);
	*(guint32 *) remote_sock->addr->address = 
	  ntohl(remote_s.sin_addr.s_addr);
	remote_sock->status = STATUS_CONNECTED;
	remote_sock->cb = sock->cb;
	remote_sock->cb_data = sock->cb_data;
	remote_sock->io_channel = g_io_channel_unix_new(remote_sock_fd);
	remote_sock->source = g_io_add_watch(remote_sock->io_channel,
					     G_IO_IN | G_IO_ERR,
					     (GIOFunc) gnome_socket_callback,
					     remote_sock);

	g_tree_insert(gnome_socket_tree,
		      &(remote_sock->tag),
		      remote_sock);

	/* call sock->cb to give it a chance to send something when 
	 * connection is ready */
	sock->cb(sock->tag,
		 GNOME_SOCKET_JUST_CONNECTED,
		 remote_sock->tag,
		 remote_sock->io_channel, /* != channel == sock->channel */
		 sock->cb_data);
      break;
      }

    default:
      g_warning("gnome_socket_callback: invalid status=%d",
		sock->status);
      break;
    }
  return TRUE;
}

static void
gnome_socket_initialize_first_time()
{
  if (gnome_socket_initialized)
    return;

  next_tag = 1;
  gnome_socket_initialized = TRUE;
  gnome_socket_tree = g_tree_new(gnome_socket_compare);
}

static gint 
gnome_socket_compare (gconstpointer one, gconstpointer two)
{
  return (*(guint *) one) - (*(guint *) two);
}

static gboolean
gnome_socket_proto_translate(GnomeNetProtocol proto,
			     gint *domain, gint *type, gint *protocol)
{
  if (domain) *domain = PF_INET; /* ony IPv4 for now */

  switch (proto)
    {
    case GNOME_NET_PROTOCOL_TCP:
      if (type) *type = SOCK_STREAM;
      if (protocol) *protocol = IPPROTO_TCP;
      break;
    case GNOME_NET_PROTOCOL_UDP:
      if (type) *type = SOCK_DGRAM;
      if (protocol) *protocol = IPPROTO_UDP;
      break;
    default:
      g_warning("gnome_socket_proto_translate: invalid protocol");
      return FALSE;
    }
  return TRUE;
}
