/* 
 * Copyright (C) 1999-2000 Roberto Zunino
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <gtk/gtk.h>
#include <string.h>

#include <config.h>

#include "gnome-string-socket.h"

#define MIN_BUF_LEN 100

typedef enum {
  STATUS_CONNECTING,
  STATUS_NORMAL,
  STATUS_REPORTING,
  STATUS_CLOSING
} Status;

typedef struct {
  Status status;
  gchar *str;
  guint size;        /* allocated memory */
  guint start, len;  /* the valid range is [start .. start+len) */
  gchar *delim;
  /* user data */
  GnomeStringSocketCallback cb;
  gpointer cb_data;
} GnomeStringSocketStruct; 

static void 
gnome_string_socket_callback(guint tag, 
			     GnomeSocketInfo info,
			     gint extra_info,
			     GIOChannel *channel,
			     gpointer data);

/* user visible functions */

guint
gnome_string_socket_connect (const gchar *hostname,
			     GnomeNetAddressType addr_type,
			     guint port,
			     const gchar *delim,
			     GnomeStringSocketCallback callback,
			     gpointer callback_data)
{
  guint tag;
  GnomeStringSocketStruct *s = NULL;

  g_return_val_if_fail(callback, 0);

  s = g_new(GnomeStringSocketStruct, 1);
  s->status = STATUS_CONNECTING;
  s->size = MIN_BUF_LEN; /* default */
  s->str = g_new(gchar, s->size);
  s->str[0] = '\0';
  s->start = s->len = 0;
  s->delim = g_strdup(delim);
  s->cb = callback;
  s->cb_data = callback_data;

  tag = gnome_socket_connect(hostname,
			     addr_type,
			     GNOME_NET_PROTOCOL_TCP, port,
			     gnome_string_socket_callback,
			     s);

  if (tag == 0) /* error from gnome_socket_connect */
    {
      g_free(s->str);
      g_free(s->delim);
      s->str = s->delim = NULL;
      g_free(s);
      return 0;
    }
  
  return tag;
}

guint
gnome_string_socket_listen(guint port, const gchar *delim,
			   GnomeStringSocketCallback callback,
			   gpointer callback_data,
			   GnomeSocketInfo *info,
			   gint *error_no)
{
  guint tag;
  GnomeStringSocketStruct *s = NULL;

  g_return_val_if_fail(callback, 0);


  s = g_new(GnomeStringSocketStruct, 1);
  s->status = STATUS_NORMAL;
  s->size = 0;
  s->str = NULL;
  s->start = s->len = 0;
  s->delim = g_strdup(delim);
  s->cb = callback;
  s->cb_data = callback_data;

  tag = gnome_socket_listen(GNOME_NET_PROTOCOL_TCP, port,
			     gnome_string_socket_callback,
			     s, info, error_no);

  if (tag == 0) /* error from gnome_socket_listen */
    {
      g_free(s->delim);
      s->delim = NULL;
      g_free(s);
      return 0;
    }
  
  return tag;
}

gboolean
gnome_string_socket_close(guint tag)
{
  GnomeStringSocketStruct *s = NULL;

  s = gnome_socket_get_callback_data(tag);
  if (s == NULL)
    return FALSE;

  switch (s->status)
    {
    case STATUS_REPORTING:
      s->status = STATUS_CLOSING;
      return TRUE;
    case STATUS_CLOSING:
      g_warning("gnome_string_socket_close:"
		" closing a closed socket (tag=%d)!\n", tag);
      return FALSE;
    case STATUS_CONNECTING:
    case STATUS_NORMAL:
      break;
    default:
      g_warning("gnome_string_socket_close: invalid status = %d\n",
		s->status);
      return FALSE;
    }

  gnome_socket_close(tag);

  if (s->str)
    {
      g_free(s->str);
      s->str = NULL;
    }
  if (s->delim)
    {
      g_free(s->delim);
      s->delim = NULL;
    }
  g_free(s);

  return TRUE;
}

/* ---------------------------------- */

static void 
gnome_string_socket_callback(guint tag, 
			     GnomeSocketInfo info,
			     gint extra_info,
			     GIOChannel *channel,
			     gpointer data)
{
  GnomeStringSocketStruct *s = (GnomeStringSocketStruct *) data;
  gint available = s->size - (s->start + s->len);
  gint len = 0, block_len, error;

  if (info == GNOME_SOCKET_JUST_CONNECTED &&
      s->status != STATUS_CONNECTING) /* listening */
    {
      GnomeStringSocketStruct *news = g_new(GnomeStringSocketStruct, 1);
      
      news->status = STATUS_NORMAL;
      news->size = MIN_BUF_LEN; /* default */
      news->str = g_new(gchar, news->size);
      news->str[0] = '\0';
      news->start = news->len = 0;
      /* for delim ref counting could be done as well... */
      news->delim = g_strdup(s->delim); 
      news->cb = s->cb;
      news->cb_data = s->cb_data;
      
      gnome_socket_set_callback_data(extra_info /* the new tag */, news);
      
      news->cb(tag, info, extra_info, channel, NULL, news->cb_data);
      return;
    }
  if (info == GNOME_SOCKET_JUST_CONNECTED &&
      s->status == STATUS_CONNECTING)
    s->status = STATUS_NORMAL;

  if (info != GNOME_SOCKET_OK) /* this handles STATUS_CONNECTING too */
    {
      s->cb(tag, info, extra_info, channel, NULL, s->cb_data);
      return;
    }
  
  if (available < s->size/10) /* low water mark */
    {
      if ((s->size - s->len) > (s->size/2)) /* shift back */
	  memmove(s->str, s->str + s->start, s->len);
      else
	{ /* increasing realloc */
	  gpointer newstr = NULL;
	  
	  s->size *= 2;
	  newstr = g_new(gchar, s->size);
	  memmove(newstr, s->str + s->start, s->len);
	  g_free(s->str);
	  s->str = newstr;
	}

      s->start = 0;
      available = s->size - s->len;
    }
  else if (s->size > MIN_BUF_LEN && 
	   available > (9*s->size)/10) /* high water mark */
    { /* decreasing realloc */
      gpointer newstr = NULL;
	  
      s->size = MAX(s->size/2, MIN_BUF_LEN); 
      /* available > 9/10 size                           -> 
       * (size - len) >= available > 9/10 size >= size/2   ->
       * -len > -size/2   ->    len < size/2 
       * therefore this is enough to contain the data 
       */
      newstr = g_new(gchar, s->size);
      memmove(newstr, s->str + s->start, s->len);
      g_free(s->str);
      s->str = newstr;
      s->start = 0;
      available = s->size - s->len;
    }

  /* sanity check */
  if (available <= 1)
    {
      g_warning("gnome_string_socket_callback: available=%d !\n"
		"Please report this bug to the gnome-socket library"
		"maintainers", available);
      return;
    }

  error = g_io_channel_read(channel,
			    s->str + s->start + s->len,
			    available - 1 /* for the '\0' */,
			    &len);
  
  if (error != G_IO_ERROR_NONE && error != G_IO_ERROR_AGAIN)
    {
      s->cb(tag, GNOME_SOCKET_READ_ERROR, error, channel, NULL, s->cb_data);
      return;
    }
  if (len == 0)
    { /* comm closed */
      /* FIXME : what about s->str ? */
      s->cb(tag, GNOME_SOCKET_CLOSED, 0, channel, NULL, s->cb_data);
      return;
    }

  s->status = STATUS_REPORTING;
  s->len += len;
  s->str[s->start + s->len] = '\0';

  block_len = strcspn(s->str + s->start, s->delim);

  while (block_len != s->len)
    {
      /* message ready */
      s->str[s->start + block_len] = '\0';
      s->cb(tag, info, extra_info, channel, s->str + s->start, s->cb_data);
      if (s->status == STATUS_CLOSING)
	{
	  s->status = STATUS_NORMAL;
	  gnome_string_socket_close(tag);
	  return;
	}
      s->start += block_len + 1 /* for the '\0' */;
      s->len -= block_len   + 1 /* for the '\0' */;
      block_len = strcspn(s->str + s->start, s->delim);
    }
  s->status = STATUS_NORMAL;
}
